library yekonga_database_model;

import 'dart:io';
import 'package:flutter/cupertino.dart';
import 'package:intl/intl.dart';
import 'package:path_provider/path_provider.dart';
import 'package:sqflite/sqflite.dart' as SQL;
import 'package:sqflite_common/sqlite_api.dart' as D_SQL;
import 'package:sqflite_common_ffi/sqflite_ffi.dart';
import 'package:uuid/uuid.dart';
import 'constants.dart';

abstract class AbstractDataModel<T> {
  Map<String, dynamic> toMap();
  T fromMap(Map<String, dynamic> map);
  dynamic valueOf(String key);
}

class DB<T> {
  final T db;
  DB(this.db);
}

/// A Database.
class Database<T extends AbstractDataModel<T>> {
  final T model;
  final String table;
  final String primaryKey;
  final List<String> columns;
  final String _databaseName = 'app_database.db';
  final int _databaseVersion = Const.DATABASE_VERSION;

  // DB db;
  List<String> _validColumns = [];
  List<String> _selected = [];
  List<String> _wheres = [];
  List<dynamic> _wheresValues = [];

  Database({
    @required this.model,
    @required this.table,
    @required this.primaryKey,
    @required this.columns,
  }) {
    for (var col in columns) {
      List<String> columnData = col.split('|');
      String columnName = (columnData.length >= 1) ? columnData[0] : null;

      if (columnName != null && columnName.isNotEmpty) {
        _validColumns.add(columnName);
      }
    }
  }

  Future<DB> open() async {
    DB db;
    String databasesDir;

    try {
      databasesDir = Directory.current.path + "/" + _databaseName;
    } catch (e) {}

    if (Platform.isAndroid || Platform.isIOS) {
      var databasesPath = await SQL.getDatabasesPath();
      String path = databasesPath + '/' + _databaseName;

      try {
        await Directory(databasesPath).create(recursive: true);
      } catch (_) {}

      db = DB<SQL.Database>(
        await SQL.openDatabase(
          path,
          version: _databaseVersion,
          onCreate: (SQL.Database db, int version) {
            return onTableCreate(DB(db), version);
          },
          onUpgrade: (SQL.Database db, int oldVersion, int newVersion) async {
            List tables = await db.rawQuery(
                "SELECT name FROM sqlite_master WHERE type = 'table'");

            for (var i in tables) {
              await db.execute("DROP TABLE IF EXISTS ${i['name']}");
            }

            await onTableCreate(DB(db), newVersion);
          },
          onOpen: (SQL.Database db) async {
            await onTableCreate(DB(db), _databaseVersion);
          },
        ),
      );
    } else {
      var databaseFactory = databaseFactoryFfi;
      if (databasesDir == null) {
        databasesDir = D_SQL.inMemoryDatabasePath;
      }
      db = DB<D_SQL.Database>(
        await databaseFactory.openDatabase(
          databasesDir,
          options: D_SQL.OpenDatabaseOptions(
            version: _databaseVersion,
            onCreate: (D_SQL.Database db, int version) {
              return onTableCreate(DB(db), version);
            },
            onUpgrade:
                (D_SQL.Database db, int oldVersion, int newVersion) async {
              List tables = await db.rawQuery(
                  "SELECT name FROM sqlite_master WHERE type = 'table'");

              for (var i in tables) {
                await db.execute("DROP TABLE IF EXISTS ${i['name']}");
              }

              await onTableCreate(DB(db), newVersion);
            },
            onOpen: (D_SQL.Database db) async {
              await onTableCreate(DB(db), _databaseVersion);
            },
          ),
        ),
      );
    }

    if (db != null && db.db.isOpen) {
      try {
        await onTableCreate(db, _databaseVersion);
      } catch (e) {
        print(e.toString());
      }
    }

    return db;
  }

  onTableCreate(DB db, int version) async {
    String q = 'CREATE TABLE IF NOT EXISTS $table (';
    List _q = [];

    _q.add('id INTEGER PRIMARY KEY');
    _selected.add('id');
    if (primaryKey != 'id') {
      _q.add(primaryKey + ' VARCHAR(24) UNIQUE');
      _selected.add(primaryKey);
    }

    for (var col in columns) {
      List<String> columnData = col.split('|');
      String columnName = (columnData.length >= 1) ? columnData[0] : null;
      String columnType = (columnData.length >= 2)
          ? ((columnData[1] == 'int') ? 'INTEGER' : columnData[1])
          : 'TEXT';
      String columnNull = (columnData.length >= 3) ? columnData[2] : 'NOT NULL';

      if (columnName == null || columnName.isEmpty || columnName == primaryKey)
        continue;
      if (_selected.indexOf(columnName) < 0) {
        _selected.add(columnName);
      }
      _q.add(columnName + ' ' + columnType + ' ' + columnNull);
    }

    q += _q.join(', ');
    q += ')';

    await db.db.execute(q);
  }

  Future<int> count() async {
    var result = await rowSummary("COUNT($primaryKey)");

    return (result == null) ? 0 : result;
  }

  Future<double> sum(String column) async {
    var result = await rowSummary("SUM($column)");

    return (result == null) ? 0 : result;
  }

  Future<dynamic> max(String column) async {
    var result = await rowSummary("MAX($column)");

    return result;
  }

  Future<dynamic> min(String column) async {
    var result = await rowSummary("MIN($column)");

    return result;
  }

  Future<dynamic> rowSummary(String query) async {
    DB db = await open();
    String where = (_wheres.length > 0) ? _wheres.join(' AND ') : "1";
    List<dynamic> whereArgs = _wheresValues;
    String _query =
        "SELECT $query AS return_value FROM $table WHERE $where LIMIT 1";
    // print(_query);
    List<Map<String, dynamic>> result = await db.db.rawQuery(_query, whereArgs);

    if (result.length > 0) {
      return result.first['return_value'];
    }

    return null;
  }

  Future<List<dynamic>> listOf(String key) async {
    List<dynamic> list = [];
    _selected = [key];
    List<T> values = await findAll();

    try {
      for (T item in values) {
        if (item != null) {
          list.add(item.valueOf(key));
        }
      }
    } catch (e) {}

    return list;
  }

  Future<List<T>> findAll() async {
    DB db = await open();
    List<T> result = List<Map<String, dynamic>>.from(
      await db.db.query(
        table,
        columns: _selected,
        where: (_wheres.length > 0) ? _wheres.join(' AND ') : ' 1 ',
        whereArgs: _wheresValues,
      ),
    ).map((e) => model.fromMap(e)).toList();

    return result;
  }

  Future<List<T>> paginate({int page: 1, int limit: 100}) async {
    DB db = await open();
    List<T> result = List<Map<String, dynamic>>.from(
      await db.db.query(
        table,
        columns: _selected,
        where: (_wheres.length > 0) ? _wheres.join(' AND ') : ' 1 ',
        whereArgs: _wheresValues,
        limit: limit,
        offset: (page - 1) * limit,
      ),
    ).map((e) => model.fromMap(e)).toList();

    return result;
  }

  Future<List<T>> find() async {
    return await this.findAll();
  }

  Future<T> findOne({String id}) async {
    DB db = await open();

    List<T> result = List<Map<String, dynamic>>.from(
      await db.db.query(
        table,
        columns: _selected,
        where: (id == null) ? _wheres.join(' AND ') : (primaryKey + ' = ?'),
        whereArgs: (id == null) ? _wheresValues : [id],
        limit: 1,
      ),
    ).map((e) => model.fromMap(e)).toList();
    // await db.close();

    if (result.length > 0) {
      return result.first;
    }

    return null;
  }

  Future<Map<String, dynamic>> insert(Map<String, dynamic> model) async {
    DB db = await open();

    if (model.containsKey(primaryKey) && model[primaryKey] == null) {
      model[primaryKey] = Uuid().v1();
    }

    Map<String, dynamic> _model = getValidValues(model);
    model[primaryKey] = await db.db.insert(table, _model,
        conflictAlgorithm: SQL.ConflictAlgorithm.replace);
    // await db.close();

    return model;
  }

  Future<List<Map<String, dynamic>>> insertAll(
      List<Map<String, dynamic>> modelList) async {
    DB db = await open();
    List<Map<String, dynamic>> list = [];

    SQL.Batch batch = db.db.batch();
    for (var model in modelList) {
      if (model.containsKey(primaryKey) && model[primaryKey] == null) {
        model[primaryKey] = Uuid().v1();
      }

      Map<String, dynamic> _model = getValidValues(model);
      batch.insert(table, _model,
          conflictAlgorithm: SQL.ConflictAlgorithm.replace);
    }
    await batch.commit(exclusive: true, noResult: true, continueOnError: true);
    // await db.close();

    // return results.map((i)=>Map<String, dynamic>.from(i)).toList();
    return list;
  }

  Future<int> update(Map<String, dynamic> model) async {
    DB db = await open();
    Map<String, dynamic> _model = getValidValues(model);

    int result = await db.db.update(
      table,
      _model,
      where: primaryKey + ' = ?',
      whereArgs: [model[primaryKey]],
      conflictAlgorithm: SQL.ConflictAlgorithm.replace,
    );
    // await db.close();

    return result;
  }

  Future<int> updateAll(List<Map<String, dynamic>> modelList) async {
    int result = 0;
    DB db = await open();
    SQL.Batch batch = db.db.batch();

    for (var model in modelList) {
      if (model.containsKey(primaryKey) && model[primaryKey] == null) {
        // model[primaryKey] = Uuid().v1();
      }

      Map<String, dynamic> _model = getValidValues(model);

      batch.update(
        table,
        _model,
        where: primaryKey + ' = ?',
        whereArgs: [model[primaryKey]],
        conflictAlgorithm: SQL.ConflictAlgorithm.replace,
      );
    }
    await batch.commit(exclusive: true, noResult: true, continueOnError: true);
    // await db.close();

    return result;
  }

  Future<int> delete({String id}) async {
    DB db = await open();
    int result;
    if (id == null && _wheres.length == 0) {
      result = await db.db.delete(table, where: '1', whereArgs: null);
    } else {
      result = await db.db.delete(table,
          where: (id == null) ? _wheres.join(' AND ') : (primaryKey + ' = ?'),
          whereArgs: (id == null) ? _wheresValues : [id]);
    }

    // await db.close();

    return result;
  }

  Future<int> deleteAll() async {
    DB db = await open();
    int result = await db.db.delete(table, where: '?', whereArgs: ['1']);

    // await db.close();

    return result;
  }

  Future<SQL.Batch> batch() async {
    DB db = await open();
    return db.db.batch();
  }

  Database whereGroup(Function callback) {
    String group = callback(this);

    return this;
  }

  Database where(String key, String oparation, dynamic value) {
    _wheres.add(key + ' ' + oparation + ' ' + '?');
    _wheresValues.add(value);

    return this;
  }

  Database orWhere(String key, String oparation, dynamic value) {
    _wheres.add('' + key + ' ' + oparation + ' ' + '?');
    _wheresValues.add(value);

    return this;
  }

  Database whereIn(String key, List<dynamic> value) {
    List<String> v = [];

    for (var item in value) {
      v.add('?');
      _wheresValues.add(item);
    }
    _wheres.add(key + ' IN ' + "(${v.join(',')})");

    return this;
  }

  Database whereNotIn(String key, List<dynamic> value) {
    List<String> v = [];

    for (var item in value) {
      v.add('?');
      _wheresValues.add(item);
    }
    _wheres.add(key + ' NOT IN ' + "(${v.join(',')})");

    return this;
  }

  Database whereDate(String key, String oparation, String value) {
    _wheres.add('Date(' + key + ') ' + oparation + ' ' + '?');
    String date = Database.formatTime(value, format: 'yyyy-MM-dd');
    _wheresValues.add(date);

    return this;
  }

  Database whereBetween(String key, String valueA, String valueB) {
    _wheres.add(key + ' BETWEEN ? AND ?');
    _wheresValues.add(valueA);
    _wheresValues.add(valueB);

    return this;
  }

  Database whereRaw(value) {
    _wheres.add(value);

    return this;
  }

  Map<String, dynamic> getValidValues(Map<String, dynamic> map) {
    Map<String, dynamic> data = {};

    map.forEach((key, value) {
      if (_validColumns.contains(key)) {
        data[key] = value;
      }
    });

    return data;
  }

  static String formatTime(dynamic date, {format: 'dd MMM, yy'}) {
    String value = '';
    DateTime time;

    try {
      if (date == null) {
        time = DateTime.now();
      } else if (date is String) {
        time = DateTime.parse(date);
      } else if (date is DateTime) {
        time = date;
      }

      value = DateFormat(format).format(time);
    } catch (e) {}

    return value;
  }
}

